/*!
 * @author Cindy Rubio Gonzalez
 */

#include "ErrorPropagation.hpp"
#include "fdd.h"
#include <algorithm>
#include "Variables.hpp"
#include <sstream>

extern int numberVars;
extern bdd identity;
extern bdd globalsToIdentity;
extern bdd allToUninitialized;
extern bdd localsToUninitialized;
extern map<int, int> mapRealToTentative;

ErrorPropagation::ErrorPropagation() : BDD(bddfalse), line_number(0), file_name(""), is_trusted(false) {}

ErrorPropagation::ErrorPropagation( bdd b) : BDD(b), line_number(0), file_name(""), is_trusted(false) {}

ErrorPropagation::ErrorPropagation( bdd b, int l, string f, bool t, string fn) : BDD(b), line_number(l), file_name(f), is_trusted(t), function_name(fn) {}

ErrorPropagation::ErrorPropagation(sem_elem_t s) {
  ErrorPropagation* t = static_cast<ErrorPropagation*>(s.get_ptr());
  BDD = t->BDD;
  line_number = t->line_number;
  file_name = t->file_name;
}


ErrorPropagation::~ErrorPropagation() {
}

string ErrorPropagation::getFunction() {
  return function_name;
}

sem_elem_t ErrorPropagation::one() const {
  bdd ONE = identity;
  return new ErrorPropagation(ONE);
}


sem_elem_t ErrorPropagation::zero() const {
  return new ErrorPropagation(bddfalse);
}


sem_elem_t ErrorPropagation::extend( SemElem* se ) {

  ErrorPropagation* rhs = static_cast< ErrorPropagation* >(se);
  
  // 1. Shifting down bdd BDD1 dom 0 -> dom 1, dom 1 -> dom 2 
  bddPair *p = bdd_newpair();
  int i;
  for(i = 0;  i < 2; i++) {
	fdd_setpair(p, i, i+1); 
  }
   
  bdd bdd1 = bdd_replace(BDD, p);

  // 2. Performing relational product 
  bdd result = bdd_relprod(bdd1, rhs->BDD, fdd_ithset(1));

  // 3. Shifting up dom 2 -> dom 1 in result 
  bddPair *p2 = bdd_newpair();
  fdd_setpair(p2, 2, 1);

  return new ErrorPropagation( bdd_replace(result, p2) );
}



sem_elem_t ErrorPropagation::combine( SemElem* se ) {
  ErrorPropagation* rhs = static_cast< ErrorPropagation* >(se);

  return new ErrorPropagation( bdd_or(BDD, rhs->BDD) );
}


bool ErrorPropagation::equal( SemElem* se ) const {
    ErrorPropagation* rhs = static_cast< ErrorPropagation* >(se);
	return BDD == rhs->BDD;
}


std::ostream & ErrorPropagation::print( std::ostream & o ) const {

  if (line_number != 0) {
	o << std::endl << "BDD: " << fddset << BDD << std::endl;
	o << "FILE: " << file_name << std::endl;
	o << "LINE: " << line_number << std::endl;

	if (is_trusted)
	  o << "TRUSTED." << std::endl;
	}
  return o;
}


sem_elem_t ErrorPropagation::getWeight( std::string& s ) {
  if (s == "ONE")
	return one();
  else
	return zero();
}

bdd ErrorPropagation::extend(const bdd &BDD1, const bdd &BDD2) {
   /* 1. Shifting down bdd BDD1 dom 0 -> dom 1, dom 1 -> dom 2 */
   bddPair *p = bdd_newpair();
   int i;
   for(i = 0;  i < 2; i++) {
     fdd_setpair(p, i, i+1); 
   }
   bdd bdd1 = bdd_replace(BDD1, p);

   /* 2. Performing relational product */
   bdd result = bdd_relprod(bdd1, BDD2, fdd_ithset(1));

   /* 3. Shifting up dom 2 -> dom 1 in result */
   bddPair *p2 = bdd_newpair();
   fdd_setpair(p2, 2, 1);

   return bdd_replace(result, p2);
}

sem_elem_t ErrorPropagation::getWeight(queue<WeightInfo>& weightInfos, int line, string file) {

  bdd result = identity;
  bdd weight;
  bool trusted = true;
  string function = "";
  

  // Adding given variable relationships
  while (! weightInfos.empty()) {
	WeightInfo w = weightInfos.front();

	trusted = trusted && w.trusted;
	function = w.function;

	if ((w.weight1 == "identity") && (w.weight2 == "identity")) {
	  weight = identity;
	}
	else if ((w.weight1 == "identityGlobals") && (w.weight2 == "identityGlobals")) {
	  weight = globalsToIdentity | localsToUninitialized;
	}
	else if ((w.weight1 == "uninitialized") && (w.weight2 == "uninitialized")) {
	  weight = allToUninitialized;
	}

	else if ((w.weight1 != "") && (w.weight2 != "")) {

	  variables->add(w.weight1);
	  variables->add(w.weight2);
	
	  int i1 = variables->getIndex(w.weight1);
	  int i2 = variables->getIndex(w.weight2);

	  weight = identity;
	  
	  weight -= (fdd_ithvar(0, i1) & fdd_ithvar(1, i1)); //removing the mapping to itself from identity bdd
	  weight |= (fdd_ithvar(0, i1) & fdd_ithvar(1, i2)); //adding new relation
	}
	else {
	  assert(false);
	}
	
	result = extend(result, weight);
	weightInfos.pop();
  }

  return new ErrorPropagation(result, line, file, trusted, function);
}


string ErrorPropagation::receivesErrorMsg(int target, int source, int &error) {
  std::ostringstream oss;

  if (line_number != 0) {
	string name_target = variables->getName(target);
	string::size_type loc1 = name_target.find("$return", 0);
	string::size_type loc2 = name_target.find("$", 0);

	if (loc1 != string::npos) { //$return was found
	  oss << file_name << ":" << line_number << ": An unchecked error may be returned" << std::endl;
	  error = mapRealToTentative[error];
	}
	else if (loc2 != string::npos) { //a parameter exchange var was found
	  oss << this->mayHaveErrorMsg(source, error);
	}
	else { // no exchange vars involved
	  oss << file_name << ":" << line_number << ":\"" << name_target 
		  << "\"" << " receives an error from \"" << variables->getName(source) << "\"" << std::endl; 
	}
  }

  return oss.str();
}


string ErrorPropagation::mayHaveErrorMsg(int target, int &error) {
  std::ostringstream oss;

  string starget = variables->getName(target);
  string::size_type loc1 = starget.find("$return", 0);

  string temp = variables->getName(error);
  int result = 0;

  if (loc1 != string::npos && (line_number != 0)) {


	result = mapRealToTentative[error];
	if (result != 0) {
	  oss<<file_name << ":" << line_number << ": An unchecked error may be returned" << std::endl;
	  error = result;
	}

  }
  else if ((line_number != 0) && (target != error)) {
	oss << file_name << ":" << line_number << ":\"" << starget 
		<< "\"" << " may have an unchecked error" << std::endl; 
  }

  return oss.str();
}


bool ErrorPropagation::hasError(const int var, const int error) {

  /* 1. Var assigned error? */
  bdd var_with_error = bddfalse;
  if (var != error) {
	bdd var_to_error = (fdd_ithvar(0,var) & fdd_ithvar(1,error));
	var_with_error = BDD & var_to_error;
  }

  if (var_with_error != bddfalse) {
	return true;
  }
  return false;
}


string ErrorPropagation::hasError(int &var, int &error, bool& slice) {
  string errorMsg = "";

  /* 1. Var assigned error? */
  if (hasError(var, error)) {
	errorMsg = receivesErrorMsg(var, error, error);
	var = error;
  }
  else {
  
	// 2. Var assigned another var?
	bdd vars_assigned = fdd_ithvar(0, var) & bddtrue; //var assigned to all vars

	int idxOK = variables->getIndex("OK");
	int idxNoVal = variables->getIndex("UNINITIALIZED");

	bdd to_remove = (fdd_ithvar(0, var) & fdd_ithvar(1, var)) |      
	  /*(fdd_ithvar(0, var) & fdd_ithvar(1, error)) | */
	  (fdd_ithvar(0, var) & fdd_ithvar(1, idxOK)) | 
	  (fdd_ithvar(0, var) & fdd_ithvar(1, idxNoVal));

	vars_assigned = vars_assigned - to_remove;

	// new
	bdd var_to_errors = bddfalse;
	vector<string>::iterator it = errorNames.begin();
	for(; it != errorNames.end(); it++) {
	  int index = variables->getIndex(*it);
	  var_to_errors |= (fdd_ithvar(0, var) & fdd_ithvar(1, index));
	}
	vars_assigned = vars_assigned - var_to_errors; 
	//until here

	
	vars_assigned = BDD & vars_assigned;

	if (vars_assigned != bddfalse) { 
	  //Finding variables assigned to var
	  int* vars = fdd_scanallvar(vars_assigned);
	  errorMsg = receivesErrorMsg(var, vars[1], error);
	  var = vars[1];
	  free(vars);
	}
	else {
      //3. The var is mapped to itself
	  slice = false;

	  string name_var = variables->getName(var);
	  string::size_type loc1 = name_var.find("$return", 0);

	  if (loc1 != string::npos)
		slice = true;
	  errorMsg = mayHaveErrorMsg(var, error);
	}
  }

  return errorMsg;

}


void ErrorPropagation::printDot(ofstream& fout ATTR_UNUSED, ofstream& foutAll ATTR_UNUSED) const {
  //std::cerr << "DOT: ErrorPropagation -- functEion: " << function_name << std::endl;
  
}


/* Yo, Emacs!
   ;;; Local Variables: ***
   ;;; tab-width: 4 ***
   ;;; End: ***
*/
