/*!
 * @author Nick Kidd/ modified by Cindy
 */

#include "FindPathVisitor.hpp"
#include "wali/witness/Witness.hpp"
#include "wali/witness/WitnessExtend.hpp"
#include "wali/witness/WitnessCombine.hpp"
#include "wali/witness/WitnessRule.hpp"
#include "wali/witness/WitnessTrans.hpp"
#include "wali/witness/WitnessMerge.hpp"
#include "Variables.hpp"
#include <typeinfo>
#include <stack>

extern bdd identity;
extern bool printforward;

typedef wali::witness::witness_t witness_t;
typedef std::list< wali::witness::witness_t > witness_ls_t;

stack<string> callstack;

FindPathVisitor::FindPathVisitor( int v, int e ) :
  var(v),original(v),error(e) {}

FindPathVisitor::~FindPathVisitor() {}

/*!
 * @return true to continue visiting children, false to stop
 */
bool FindPathVisitor::visit( wali::witness::Witness * w ATTR_UNUSED) {
  return true;
}


/*!
 * @return true to continue visiting children, false to stop.
 */
bool FindPathVisitor::visitExtend( wali::witness::WitnessExtend * w ) {

  if (var == error) {
	std::cerr << "[VisitExtend] Stopping...\n";
	return false;
  }

  if( w->hasLeft() && w->hasRight() ) {
	witness_t l = w->left();
	witness_t r = w->right();

	r->accept(*this);

	if (var != error) {
	  l->accept(*this);
	}

  }
  else {
	// does not have 2 children so just go down whichever path
	return true;
  }
  
  return false;
}

  
/*!
 * @return true to continue visiting children, false to stop.
 */
bool FindPathVisitor::visitCombine( wali::witness::WitnessCombine * w ) {

  ErrorPropagation* weight = witnessToWeight(w);

  //if (weight->hasError(var, error, false, false, false)) { //has error without switch? 
  if (weight->hasError(var, error)) { 
	witness_ls_t& ls = w->children();
	witness_ls_t::iterator it = ls.begin();
	witness_ls_t::iterator itEND = ls.end();

	for( ; it != itEND ; it++ ) {
	  witness_t combChild = *it;
	  ErrorPropagation *weightChild = witnessToWeight(combChild.get_ptr());

	  // is this the child we are looking for?
	  //if( weightChild->hasError(var, error, false, false, false) ) { //has error without switch?
      if (weightChild->hasError(var, error)) {
		combChild->accept(*this);
		break;
	  }
	}
  }
  // we are handling recursing ourself
  return false;
}


/*!
 * @return true to continue visiting children, false to stop.
 */
bool FindPathVisitor::visitRule( wali::witness::WitnessRule * w ) {

  ErrorPropagation* weight = witnessToWeight(w);

  // Adding to path...
  bool slice = true;
  string msg = weight->hasError(var, error, slice);

  if (msg != "") {
	pathMsgs.push_back(msg);
	if (slice) sliceMsgs.push_back(msg);
	pathWeights.push_back(weight);
  }

  return false;
}


/*!
 * @return true to continue visiting children, false to stop.
 */
bool FindPathVisitor::visitTrans( wali::witness::WitnessTrans * w ATTR_UNUSED) {
  return true;
}


bool FindPathVisitor::visitMerge( wali::witness::WitnessMerge * w ) {

  if( w->hasLeft() && w->hasRight() ) {
	witness_t l = w->left();
	witness_t r = w->right();
	
	r->accept(*this);
	
	if (var != error) {
	  l->accept(*this);
	}
  }
  else {
	// does not have 2 children so just go down whichever path
	return true;
  }
  
  return false;
}


ErrorPropagation* FindPathVisitor::witnessToWeight( wali::witness::Witness* w ) {

  ErrorPropagation* weight = static_cast<ErrorPropagation*>( w->weight().get_ptr() );
  if( 0 == weight ) {
	std::cerr << "[? ERROR ?] FindPathVisitor::visitExtend\n";
	assert(0);
	return false;
  }
  return weight;
}



std::ostream& FindPathVisitor::printPath( std::ostream& out, string file, int line, string var, string msgcode, bool tentative ) {
  
  /*Printing summary function names*/
  /*For now just when printing in forward direction, so that I do not have to replace expected output files for tests*/
  
 if (printforward) {

   out << "SUMMARY:";

   if (pathMsgs.size() > 0) {
	 string firstMsg = pathMsgs.back();
	 
	 string::size_type loc = firstMsg.find("mm");
	 if (loc != string::npos) {
	   out << "mm:";
	 }
	 else {
	   out << "fs:";
	 }
   }
   else {
	 out << ":";
   }
   
   out << msgcode << ":";
	
   ErrorPropagation* firstWeight = pathWeights.back();
   ErrorPropagation* lastWeight = pathWeights.front();
   if (firstWeight != NULL && lastWeight != NULL) {
	 
	 string firstFunction = firstWeight->getFunction();
	 string lastFunction = lastWeight->getFunction();
	 
	 out << firstFunction << ":" << lastFunction << "\n";
   }
   else {
	 out << ":\n";
   } 
 }
  
 /*Printing path information*/
 
 
 string msg;
 if (msgcode == "overwrite") {
   if (tentative)
	 msg =  ": overwriting potential unchecked error in \"";
   else
	 msg =  ": overwriting potential non-tentative unchecked error in \"";
 }
 else {	
   if (tentative)
	 msg = ": potential unchecked error in out of scope variable \"";
   else 
	 msg = ": potential non-tentative unchecked error in out of scope variable \"";
 }


  if (printforward) {
	std::list<string>::reverse_iterator it;
	for( it = pathMsgs.rbegin(); it != pathMsgs.rend(); it++ ) {
	  out << *it; 
	}
	out << file << ":" << line << msg << var << "\"\n";
  }

  else {
	out << file << ":" << line << msg << var << "\"\n";

	std::list<string>::iterator it;
	for( it = pathMsgs.begin(); it != pathMsgs.end(); it++ ) {
	  out << *it; 
	}
  }
  
  return out;
 }


std::ostream& FindPathVisitor::printSlice( std::ostream& out, string file, int line, string var, string msgcode, bool tentative ) {

  string msg;
  if (msgcode == "overwrite") {
	if (tentative)
	  msg =  ": overwriting potential unchecked error in \"";
	else
	  msg =  ": overwriting potential non-tentative unchecked error in \"";
  }
  else {	
	if (tentative)
	  msg = ": potential unchecked error in out of scope variable \"";
	else 
	  msg = ": potential non-tentative unchecked error in out of scope variable \"";
  }



  if (printforward) {
	std::list<string>::reverse_iterator it;
	for(it = sliceMsgs.rbegin(); it != sliceMsgs.rend(); it++ ) {
	  out << *it;
	}
	out << file << ":" << line << msg << var << "\"\n";
  }
  else {
	out << file << ":" << line << msg << var << "\"\n";
	
	std::list<string>::iterator it;
	for(it = sliceMsgs.begin(); it != sliceMsgs.end(); it++ ) {
	  out << *it;
	}
  }

  return out;
}

std::ostream& FindPathVisitor::printWeights( std::ostream& o ) {
  std::list<ErrorPropagation*>::iterator it;
  for(it = pathWeights.begin(); it != pathWeights.end(); it++ ) {
	(*it)->print(o);
  }
  return o;
}


void FindPathVisitor::printDot( ofstream& foutAll, ofstream& fout, string overwritingFunction) {
  
  fout << "digraph G {" << endl;
  fout << "  rankdir=\"TB\";" << endl;
  fout << "  page=\"8.5,11\";" << endl;
  fout << "  orientation=\"landscape\";" << endl;
  fout << "  size=\"20,20\";" << endl;
  
  callstack.empty();
  callstack.push(overwritingFunction);

  // Drawing first node and determining its color
  // Orange and source and target are the same
  // Otherwise green

  //path in reverse order, thus 'origin' weight is at the end of the list
  ErrorPropagation* originWeight = pathWeights.back(); 
  string originFunction = originWeight->getFunction();
  /*delete*///std::cerr << "overwriting: " << overwritingFunction << "    original: " 
			//<< originFunction << std::endl;

  string originColor = "yellow";
  string overwritingColor = "red";

  if (overwritingFunction == originFunction) {
	overwritingColor = originColor = "orange";
  }


  fout << "  " << overwritingFunction << "[ fillcolor=\"" << overwritingColor 
	   << "\" style=\"filled\" label=\"" << overwritingFunction << "\"];" << endl;
  foutAll << "  " << overwritingFunction << "[ fillcolor=\"" << overwritingColor 
		  << "\" style=\"filled\" label=\"" << overwritingFunction << "\"];" << endl;

  std::list<ErrorPropagation*>::reverse_iterator it;
  for(it = pathWeights.rbegin(); it != pathWeights.rend(); it++ ) {
	(*it)->printDot(fout, foutAll);
  }

  //Drawing last node again ...
  fout << "  " << originFunction << "[ fillcolor=\"" << originColor 
  	   << "\" style=\"filled\" label=\"" << originFunction << "\"];" << endl;
  foutAll << "  " << originFunction << "[ fillcolor=\"" << originColor 
  	   << "\" style=\"filled\" label=\"" << originFunction << "\"];" << endl;


  fout << "}" << endl;

  return;
}

/* Yo, Emacs!
   ;;; Local Variables: ***
   ;;; tab-width: 4 ***
   ;;; End: ***
 */
