
/*
 * @author Nick Kidd, Modified by Cindy Rubio Gonzalez
 */

#include "wali/QueryHandler.hpp"
#include "wali/wpds/WpdsHandler.hpp"
#include "wali/wfa/WfaHandler.hpp"
#include "StrX.hpp"
#include "ErrorPropagation.hpp"
#include "MergeError.hpp"
#include "Variables.hpp"
#include <map>
#include <memory>
#include <set>

#include "ParseArgv.hpp"
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/sax2/SAX2XMLReader.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>
#include <xercesc/util/OutOfMemoryException.hpp>
#include "fdd.h"
#include "wali/wpds/ewpds/EWpdsHandler.hpp"
#include "wali/wpds/ewpds/EWPDS.hpp"
#include "wali/wpds/fwpds/FWPDS.hpp"        
#include "wali/witness/WitnessWrapper.hpp"
#include "wali/wfa/State.hpp"
#include "wali/witness/Witness.hpp"
#include "FindPathVisitor.hpp"
#include "ProgramPoint.hpp"


XERCES_CPP_NAMESPACE_USE

typedef vector<ProgramPoint> ProgramPoints;
extern ProgramPoints programPoints;
extern vector<string> globalVars;

bool printforward = false;
bool printdot = false;
string dotfile;
extern vector<string> errorNames;

//bdd bddErrors;
//bdd bddErrorsOK;
//map<int, bdd> mapErrorToBdd;  
//bdd bddTentativeErrors;
//map<int, int> mapTentativeToReal;


int parseWpds( std::string& xmlFile );
int parseWfa( std::string& xmlFile );
int parseQuery( std::string& xmlFile );
int parseFile( DefaultHandler&, std::string& xmlFile );
//void populatingErrorDS();
void openFiles(ofstream& foutAll, ofstream& fout);
void closeFiles(ofstream& foutAll, ofstream& fout);


int main( int argc, char** argv ) {

  k::ParseArgv parser(argc,argv);
  std::string fname;

  if (parser.get("--error-codes", fname))
    variables.reset(new Variables(fname));
  else
    {
      cerr << "missing required \"--error-codes\" option\n";
      return 2;
    }

  int rc=0;
  if( parser.get("--printforward",fname) ) {
    printforward = true;
  }
  if (parser.get("--printdot", dotfile)) {
    printdot = true;
  }

  if( parser.get("--wpds",fname) ) {
    rc |= parseWpds(fname);
  }
  if( parser.get("--wfa",fname) ) {
    rc |= parseWfa(fname);
  }
  if( parser.get("--query",fname) ) {
    rc |= parseQuery(fname);
  }

  return rc;
}

int parseWpds( std::string& xmlFile )
{
  
  // using wpds (1)
  /*
    ErrorPropagation wf(bddfalse);   
    wali::wpds::WpdsHandler handler(wf);                                  
  */
  
  // using ewpds    
  wali::wpds::fwpds::FWPDS *fwpds = new wali::wpds::fwpds::FWPDS(); 
  ErrorPropagation wf(bddfalse);   
  MergeError mf("");
  wali::wpds::ewpds::EWpdsHandler handler(fwpds, wf, mf);       
  // until here
  
  int rc = parseFile(handler,xmlFile);
  handler.get().print( std::cout );
  return rc;
}

int parseWfa( std::string& xmlFile )
{
  ErrorPropagation wf(bddfalse);
  wali::wfa::WfaHandler handler(wf);
  int rc = parseFile(handler,xmlFile);
  handler.get().print( std::cout );
  return rc;
}

int parseQuery( std::string& xmlFile )
{

  // using wpds (2)
  /*
  ErrorPropagation wf(bddfalse);
  wali::QueryHandler handler(wf);
  */

  // using ewpds
  wali::witness::WitnessWrapper wrapper;

  wali::wpds::fwpds::FWPDS *fwpds = new wali::wpds::fwpds::FWPDS(&wrapper);
  ErrorPropagation wf(bddfalse);
  MergeError mf("");
  
  wali::wpds::ewpds::EWpdsHandler fwpdsHandler(fwpds, wf, mf);       
  wali::wfa::WfaHandler wfaHandler(wf);                              
  
  wali::QueryHandler handler(&fwpdsHandler, &wfaHandler);            
  // until here
  
  
  
  /*****************************/
  /*   Parsing wpds XML file   */
  /****************************/
  /*delete*/ std::cerr << "1. Parsing XML file and creating rules...\n";
  
  int rc = parseFile(handler,xmlFile);
  if( 0 == rc ) {
    
    /*****************************/
    /*   Constructing WFA        */
    /****************************/  
    /*delete*/ std::cerr << "2. Constructing WFA...\n"; 
    wali::wfa::WFA result = handler.run();
    
    /*delete*/ std::cerr << "3. Inspecting interesting points in the program ... A total of " 
			 << programPoints.size() << " points were found..." << std::endl;
    
    
    
    //   populatingErrorDS();
    
    result.path_summary();
    wali::wfa::State* st = result.getState(result.getInitialState());
    
    // Populating map from trans to key so that we can retrieve any trans given a key without the need to iterate through all transtitions
    map<wali::Key, wali::wfa::Trans*> mapTrans;
    wali::wfa::State::iterator it = st->begin();
    for(; it != st->end(); it++) {
      wali::wfa::Trans* t = *it;
      wali::Key s = t->stack();
      mapTrans[s] = t;
    }
    
    
    // Iterating through potential interesting program points
    wali::wfa::TransSet transSet = st->getTransSet();
    for(unsigned int i = 0; i < programPoints.size(); i++) {
      programPoints[i].calculateWeight(result, mapTrans);
      programPoints[i].processTargets();
    }
    
    
  }
  return rc;
}



int parseFile( DefaultHandler& handler,std::string& xmlFile )
{
    int errorCode = 0;

    // Initialize the XML4C2 system
    try {    
        XMLPlatformUtils::Initialize();
    }   
    catch (const XMLException& toCatch) {       
        XERCES_STD_QUALIFIER cerr << "Error during initialization! :\n" << StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;
        return 1; 
    }

    SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();


    try {
        parser->setContentHandler(&handler);
        parser->setErrorHandler(&handler);
        parser->parse(xmlFile.c_str());
    }
    catch( const OutOfMemoryException&)
    {
        XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
        errorCode = 2;
    }
    catch (const XMLException& toCatch)
    {
        XERCES_STD_QUALIFIER cerr << "\nAn error occurred\n  Error: "
            << StrX(toCatch.getMessage())
            << "\n" << XERCES_STD_QUALIFIER endl;
        errorCode = 3;
    }
    catch( const SAXException& toCatch ) {
        std::cerr << "Error in Main: " << StrX(toCatch.getMessage()) << std::endl;
        errorCode = 4;
    }

    delete parser;

    XMLPlatformUtils::Terminate();

    return errorCode;
}

/*
void populatingErrorDS() {

  for(unsigned int i = 0; i < errorNames.size(); i++) {
    int index = variables->getIndex(errorNames[i]);
    bddErrors |= fdd_ithvar(1, index);

    //new
    int indexTentative = variables->getIndex("TENTATIVE_" + errorNames[i]);
    bddTentativeErrors |= fdd_ithvar(1, indexTentative);

    mapTentativeToReal[indexTentative] = index;
    // until here
  }


  bddErrorsOK = bddErrors | fdd_ithvar(1, variables->getIndex("OK"));
  
  for(unsigned int i = 0; i < errorNames.size(); i++) {
    int index = variables->getIndex(errorNames[i]);
    bdd result = bddErrorsOK - fdd_ithvar(1, index);
    mapErrorToBdd[index] = result;
  }
}
*/


void openFiles(ofstream& foutAll, ofstream& fout) {
    string fileAll = "all-" + dotfile;
    foutAll.open(fileAll.c_str());
    
    foutAll << "digraph G {" << endl;
    foutAll << "  rankdir=\"TB\";" << endl;
    foutAll << "  page=\"8.5,11\";" << endl;
    foutAll << "  orientation=\"landscape\";" << endl;
    foutAll << "  size=\"20,20\";" << endl;
    
    fout.open(dotfile.c_str());
}


void closeFiles(ofstream& foutAll, ofstream& fout) {
  foutAll << "}" << endl;
  foutAll.close();
  fout.close();
}

// if (printdot)
// path.printDot(foutAll, fout, function);
