#!/usr/bin/env python

import sys
import re;
import os;

from xml.sax import make_parser
from xml.sax import handler
from xml.sax.handler import feature_namespaces
from xml.sax.handler import feature_validation
#from parser import *

pathSeparator='/'
if sys.platform == 'win32':
    pathSeparator = '\\'

##################################
# Helper methods
##################################
def expand(s,env=os.environ):
    """
    expand takes as input a string and an environment and attempts
    to do shell like expansion.

    input: S                 <! a string
    input: env               <! mapping from strings to strings
    output: S such that each occurrence of $(X) in S is 
            exanded into the mapping of X in env
    """
    prev = 0
    result = list()
    first = 1
    p = re.compile('\$\((\w+)\)')
    iterator = p.finditer(s)
    for match in iterator:
        key = match.group(1)
        try:
            result.append(s[prev:match.start()])
            result.append(env[key])
        except KeyError:
            result.append(match.group())
        prev = match.end()
    # Get the rest of the string that comes after the last match
    result.append(s[prev:])
    #print result
    return ''.join(result)

##################################
# Project class definitions
##################################
class Project:
    """
    Defines Project interface and provides default implementation for some
    methods.
    """
    def __init__(self,name,path):
        # strip trailing '/'
        path = self.stripTrailingSlash(path)
        self.name = name
        self.projectdir = path
        self.builddir=path
        self.outputdir = path
        self.libtool = 'libtool'
        if sys.platform.startswith('darwin'):
            self.libtool = 'glibtool'
        self.cc = 'gcc'
        self.cflags = ''
        # Dict of include dirs 
        self.includes = dict()
        # Dict of input files
        self.inputs = dict()
        # Dict of libs needed
        self.libs = dict()
        # Dict of projects this project depends on
        self.dependencies = dict()

    def stripTrailingSlash(self,path):
        while path.endswith('/'):
            path = path[0:-1]
        return path

    def compiler(self,cc=None):
        if( cc != None):
            self.cc = cc
        return self.cc

    def buildDir(self,path=None):
        if( path != None ):
            path = self.stripTrailingSlash(path)
            self.builddir = path
        return self.builddir

    def outputDir(self,path=None):
        if( path != None):
            path = self.stripTrailingSlash(path)
            self.outputdir = path
        return self.outputdir

    def set_cflags(self,cflags):
        self.cflags = cflags

    def add_item(self,dict,name,path,is_rel):
        if is_rel:
            dict[name] = '$('+self.projectDirVar()+')/'+path
        else:
            dict[name] = path

    def add_input(self,name,path,is_rel):
        self.add_item(self.inputs,name,path,is_rel)

    def add_include(self,name,path,is_rel):
        self.add_item(self.includes,name,path,is_rel)
        self.includes[name] = '-I'+self.includes[name]

    def add_lib(self,name,path,is_rel):
        self.add_item(self.libs,name,path,is_rel)

    def addDependency( self, projName, project ):
        self.dependencies[projName] = project
        print '#Added dependency "%s" to "%s"' % (projName,self.name)

    def outputName(self):
        """ Return the name of project output"""

    def binaryName(self):
        """ Return the binary created by the project"""

    def cleanVar(self):
        """ Return the name of the Makefile 'clean' command
            for this project
        """
        return self.name+'.CLEAN'

    def ccVar(self):
        """
        Return the name of the Makefile's variable for the project's
        cc
        """
        return self.name+'.CC'

    def cflagsVar(self):
        """
        Return name of Makefile's variable for project's cflags var
        """
        return self.name+'.CFLAGS'

    def includesVar(self):
        """
        Return name of Makefile's variable for project's includes var
        """
        return self.name+'.INCLUDES'

    def libsVar(self):
        """
        Return the name of Makefile's variable for the project's lib var
        """
        return self.name+'.LIBRARIES'

    def buildDirVar(self):
        """
        Return name of Makefile's variable for project's builddir var
        """
        return self.name+'.BUILDDIR'

    def projectDirVar(self):
        """
        Return name of Makefile's variable for project's project dir var
        """
        return self.name+'.PROJECTDIR'

    def outputDirVar(self):
        """
        Return name of Makefile's variable for project's outputdir var
        """
        return self.name+'.OUTPUTDIR'

    def dumpProjectVars(self):
        """ Writes all the project vars to the output """
        print self.buildDirVar(),'=',self.buildDir()
        print self.projectDirVar(),'=',self.projectdir
        print self.outputDirVar(),'=',self.outputDir()
        print self.ccVar(),'=',self.compiler()
        print self.cflagsVar(),'=',self.cflags
        print self.includesVar(),'='," ".join(self.includes.values()),
        for dep in self.dependencies.values():
            print '$(%s)' % (dep.includesVar()),
        print
        print self.libsVar(),'=',
        for key,val in self.libs.items():
            print '-L%s -l%s' % (val,key),
        for dep in self.dependencies.values():
            if( dep.__class__.__name__ == LibProject.__name__):
                print '-L$(%s) -l%s' % (dep.outputDirVar(),dep.name),
            print '$(%s)' % (dep.libsVar()),
        print
        print

    def dump(self):
        """ Writes Project to stdout.
            Base class only knows how to build each the project
            inputs
        """
        self.dumpProjectVars()

        # inputs
        for key,val in self.inputs.items():
            print "$(%s)/%s.%s.lo: %s" % (self.buildDirVar(),self.name,key,val)
            print "\t%s --mode=compile $(%s) $(%s) $(%s) -o $@ -c $^" % (
                self.libtool
                ,self.ccVar()
                ,self.cflagsVar()
                ,self.includesVar())
            print
        # clean
        print '.PHONY:',self.cleanVar()
        print '%s:' % (self.cleanVar())
        print '\tcd $(%s);' %(self.outputDirVar()),
        print '%s --mode=clean rm -f %s; cd - ; %s --mode=clean rm -f ' % (self.libtool,self.binaryName(),self.libtool),
        for key in self.inputs.keys():
            print '$(%s)/%s.%s.lo' % (self.buildDirVar(),self.name,key),
        print
        print

###############
# ExeProject
###############
class ExeProject(Project):
    """ Defines project for building an executable """

    def outputName(self):
        """ Returns the Projects output name as a string """
        return '$(%s)/%s' % (self.outputDirVar(),self.name)

    def binaryName(self):
        return self.name

    def dump(self):
        Project.dump(self)
        print self.outputName()+':',
        for key in self.inputs.keys():
            print '$(%s)/%s.%s.lo' % (self.buildDirVar(),self.name,key),
        print
        print '\t%s --mode=link $(%s) $(%s) -avoid-version -o $@ $^ $(%s)' % (
            self.libtool
            , self.ccVar()
            , self.cflagsVar()
            , self.libsVar() )
        print
        print '.PHONY: %s' % (self.name)
        print '%s:' % (self.name),
        for dep in self.dependencies.values():
            print dep.name,
        print
        print '\t( test -d $(%s) || mkdir $(%s) ) && \\' % (self.buildDirVar(),self.buildDirVar())
        print '\t( test -d $(%s) || mkdir $(%s) ) && \\' % (self.outputDirVar(),self.outputDirVar())
        print '\t$(MAKE) %s' % (self.outputName())
        print
        print

###############
# LibProject
###############
class LibProject(Project):
    """ Defines project for building a library """

    def outputName(self):
        """ Returns the Projects output name as a string """
        return '$(%s)/lib%s.la' % (self.outputDirVar(),self.name)

    def binaryName(self):
        return 'lib%s.la' % (self.name)

    def dump(self):
        Project.dump(self)
        print self.outputName()+':',
        for key in self.inputs.keys():
            print '$(%s)/%s.%s.lo' % (self.buildDirVar(),self.name,key),
        print
        print '\t%s --mode=link $(%s) $(%s) -avoid-version -rpath %s -o $@ $^' % (
            self.libtool
            , self.ccVar()
            , self.cflagsVar()
            , self.outputDir())
        print
        print '.PHONY: %s' % (self.name)
        print "%s:" % ( self.name ),
        for dep in self.dependencies.values():
            print dep.name,
        print
        print '\t( test -d $(%s) || mkdir $(%s) ) && \\' % (self.buildDirVar(),self.buildDirVar())
        print '\t( test -d $(%s) || mkdir $(%s) ) && \\' % (self.outputDirVar(),self.outputDirVar())
        print '\t$(MAKE) %s' % (self.outputName())
        print
        print

###############
# EmptyProject
###############
class EmptyProject(Project):
    """
    An EmpyProject generates a makefile build target
    that will compile several projects. Commonly used
    for tests/examples/etc.
    """

    def outputName(self):
        return self.name

    def dump(self):
        """ Dump Makefile commands to stdout for the group """
        Project.dumpProjectVars(self)
        print '.PHONY:',self.cleanVar()
        print '%s:' % (self.cleanVar()),
        for cleanProj in self.dependencies.values():
            print cleanProj.cleanVar(),
        print
        print '.PHONY:' , self.outputName()
        print self.name+':',
        for buildProj in self.dependencies.values():
            print buildProj.name,
        print
        print
        print

##################################
# Parser class definition
##################################
class MkMfParser(handler.ContentHandler,handler.ErrorHandler):
    MkMfElementName = 'mkmf'
    ProjectElementName = 'Project'
    CompilerElementName = 'Compiler'
    IncludesElementName = 'IncludeList'
    InputsElementName = 'InputList'
    LibsElementName = 'LibraryList'
    DepsElementName = 'DependencyList'
    ProjectGroupElementName = 'ProjectGroup'
    ItemElementName = 'item'

    def __init__(self, outfile):
        self.outfile = outfile
        self.projects = dict()
        self.currentProject = None
        self.environ = dict()
        # copy the os environ
        for k,v in os.environ.items():
            self.environ[k] = v

        # MkMf settings
        self.topdir = '.'
        self.in_inputs=0
        self.in_includes=0
        self.in_libs=0
        self.in_deps=0
        self.allList = list()
        self.default=None

    def prepareForProject(self):
        if( self.environ.has_key('ProjectDir') ):
            del self.environ['ProjectDir']
        if( self.environ.has_key('BuildDir') ):
            del self.environ['BuildDir']
        if( self.environ.has_key('OutputDir') ):
            del self.environ['OutputDir']

    def startDocument(self):
        """ Begin mkmf file"""
        # header
        print "#######################################"
        print "# -- Begin MkMf generated Makefile -- #"
        print "#######################################"
        print
        # default, all, and clean rules for Makefile
        print '.PHONY: default all clean'
        print 'default: mkmf-default'
        print 'all: mkmf-all'
        print 'clean: mkmf-clean'
        print

    def endDocument(self):
        """ End mkmf file """
        # default
        if self.projects.has_key(self.default):
            print '.PHONY: mkmf-default'
            print 'mkmf-default:',self.projects[self.default].name
            print
        # all
        print '.PHONY: mkmf-all'
        print 'mkmf-all:',
        for project in self.allList:
            print project.name,
        print
        print
        #clean
        print '.PHONY: mkmf-clean'
        print 'mkmf-clean:',
        for project in self.allList:
            print project.cleanVar(),
        print
        print
        # footer
        print "#######################################"
        print "# -- End MkMf generated Makefile --   #"
        print "#######################################"


    def startElement(self, name, attrs):
        # mkmf
        if name == MkMfParser.MkMfElementName:
            if attrs.has_key('RelativePath'):
                self.topdir = os.getcwd() + '/'+attrs['RelativePath']
            elif attrs.has_key('Path'):
                self.topdir = attrs['Path']
            if attrs.has_key('Name'):
                self.mkmfName = attrs['Name'] + '.TOPDIR'
            else:
                self.mkmfName = 'mkmf' + '.TOPDIR'
            #if self.topdir.endswith('/') == 0:
            #    self.topdir = self.topdir+'/'
            self.topdir = expand(self.topdir,self.environ)
            self.environ['MkmfDir'] = self.topdir
            self.default = attrs['Default']
            # Everything that uses relative paths will be encoded
            # using the Makefile var $(mkmf.TOPDIR)
            print "#########################################"
            print "#TOPDIR: All projects with relative paths"
            print "#        will be offset from this"
            print "#########################################"
            print "%s=%s" % (self.mkmfName,self.topdir)
            print
            self.topdir="$(%s)/" % (self.mkmfName)

        # project
        if name == MkMfParser.ProjectElementName:
            self.prepareForProject()
            projType = attrs['Type']
            projName = attrs['Name']
            projRP = self.topdir

            # Check for project relative path
            if( attrs.has_key('RelativePath') ):
                projRP = self.topdir+attrs['RelativePath']
                projRP = expand(projRP,self.environ)

            if projType == 'Library':
                self.currentProject = LibProject(projName, projRP)
            elif projType == 'Executable':
                self.currentProject = ExeProject(projName,projRP)
            elif projType == 'Empty':
                self.currentProject = EmptyProject(projName,projRP)
            else:
                raise "Unsuppored Project"

            self.environ['ProjectDir'] = projRP
            if attrs.has_key('BuildDir'):
                self.environ['BuildDir'] = expand(self.topdir+attrs['BuildDir'],self.environ)
                self.currentProject.buildDir(self.environ['BuildDir'])
            if attrs.has_key('OutputDir'):
                self.environ['OutputDir'] = expand(self.topdir+attrs['OutputDir'],self.environ)
                self.currentProject.outputDir(self.environ['OutputDir'])
            self.projects[projName] = self.currentProject
        # inputs
        if name == MkMfParser.InputsElementName:
            self.in_inputs = 1
        # libs
        elif name == MkMfParser.LibsElementName:
            self.in_libs = 1
        # includes
        elif name == MkMfParser.IncludesElementName:
            self.in_includes = 1
        # deps
        elif name == MkMfParser.DepsElementName:
            self.in_deps = 1
        # compiler
        elif name == MkMfParser.CompilerElementName:
            self.handleCompiler(attrs)
        # item
        elif name == MkMfParser.ItemElementName:
            self.handleItem(attrs)

    def handleDep(self,attrs):
        depName = attrs['Name']
        if self.projects.has_key(depName):
            self.currentProject.addDependency(depName,self.projects[depName])
        else:
            sys.stderr.write('[WARNING] No Project with ref "%s" seen\n' % depName)

    def handleItem(self,attrs):
        itemName = attrs['Name']
        if attrs.has_key('RelativePath'):
            itemPath = expand(attrs['RelativePath'],self.environ)
            is_rel = 1
        elif attrs.has_key('Path'):
            itemPath = expand(attrs['Path'],self.environ)
            is_rel = 0
        if self.in_inputs:
            self.currentProject.add_input(itemName,itemPath,is_rel)
        elif self.in_libs:
            self.currentProject.add_lib(itemName,itemPath,is_rel)
        elif self.in_includes:
            self.currentProject.add_include(itemName,itemPath,is_rel)
        elif self.in_deps:
            self.handleDep(attrs)

    def handleCompiler(self, attrs):
        if attrs.has_key('Name'):
            self.currentProject.compiler(attrs['Name'])
        if attrs.has_key('cflags'):
            self.currentProject.set_cflags(attrs['cflags'])

    def endElement(self, name):                                 # [7]
        if name == MkMfParser.InputsElementName:
            self.in_inputs = 0
        elif name == MkMfParser.LibsElementName:
            self.in_libs = 0
        elif name == MkMfParser.IncludesElementName:
            self.in_includes = 0
        elif name == MkMfParser.DepsElementName:
            self.in_deps = 0
        elif name == MkMfParser.ProjectElementName:
            self.currentProject.dump()
            self.allList.append(self.currentProject)


##########################
# Actual run method
##########################
def run(filename):
    """
    Run the MkMfParser on input filename.
    Dumps Makefile to stdout
    """
    # Create a parser
    parser = make_parser()

    # Tell the parser we are not interested in XML namespaces
    #parser.setFeature(feature_namespaces, 1)
    #parser.setFeature(feature_validation, 1)

    # Create the handler
    dh = MkMfParser(sys.stdout)

    # Tell the parser to use our handler
    parser.setContentHandler(dh)

    # Parse the input
    parser.parse(filename)

def usage():
    """
    Print usage info
    """
    print """
    python mkmf.py <input.mkmf> [<output>]
    """

def main( argv = None ):
    if argv == None:
        argv = sys.argv[1:]
    if len(argv) == 1:
        run(argv[0])
        return 0
    elif len(argv) == 2:
        try:
            makefile = open(argv[1],'w')
            oldstdout = sys.stdout
            sys.stdout = makefile
            run(argv[0])
            makefile.close()
            sys.stdout = oldstdout
        except IOError:
            sys.stderr.write('Cannot open %s\n' % argv[1])
            run(argv[0])
    else:
        usage()
        return 1

if __name__ == '__main__': main()


