(*
   Extract a WPDS from C code using CIL
   @author Nicholas Kidd / modified by Cindy Rubio
*)

open Cil
open Pretty
open Ptranal
open Hashtbl
open List
open Rmtmps
open Deadcodeelim
open Cfg
open ConnectFunctionPointers
open Filters

(* Globals *)
let numberVars = ref 0
let numberGlobalVars = ref 0
let tableFuncNumArgs = Hashtbl.create 2300
let tableFuncAcls = Hashtbl.create 2300
let tableFuncFmls = Hashtbl.create 2300
let tableFuncLocals = Hashtbl.create 2300
    

(*Source visitor begins*)    
class sourceVisitor theFile ?outChan:(pdsout=stdout) ?errChan:(pdserr=stderr) () = object(self)
  inherit nopCilVisitor

  val mutable thefundec = None;
    
  method vstmt(s: stmt) = begin
    match s.skind with
    | Instr[Call(None, f, args, loc)] ->
	begin 
          match f with
          | Lval(Var(info),_) ->
              begin
		match info.vtype with
                | TFun(TVoid(_),_,_,_) -> (*it is indeed void*)
                    ignore();
		| TFun(t,_,_,_) -> (*it is not void*)  
                    begin
                      match thefundec with
                      | Some fdec ->
                          s.skind <- Instr[ Call(Some (var(makeTempVar fdec t)), f, args, loc)]
                      | None -> ignore(); (*cannot happen*)
                    end;
		| _ -> ignore();
              end;
          | Lval(Mem(memExp),oset) -> ignore(); (*do later function pointers*)
          | _ -> ignore();
	end;
    | _ -> ignore(); (*no replacements to be done*)
  end;
    DoChildren;

  method vfunc(f: fundec) =
    IsolateInstructions.visit f;
    prepareCFG f;
    computeCFGInfo f false;
    
    thefundec <- Some f;
    DoChildren;

end
(* End of source visitor *)



(* Counter visitor begins*)
class counterVisitor theFile ?outChan:(pdsout=stdout) ?errChan:(pdserr=stderr) () = object(self)
  inherit nopCilVisitor

  method createIntList(initial:int)(length:int) =
    if (length <= 0) then
      []
    else
      initial::(self#createIntList (initial + 1) (length - 1));
    
  method doGlobal(name:string) = 
    numberVars := !numberVars + 1;
    numberGlobalVars := !numberGlobalVars + 1;
    ignore( fprintf pdsout "\t\t<var name=\"%s\"/>\n" name)

  method vvdec(v:varinfo) =
    begin match v.vtype with
    | TFun(_,_,_,_) ->
        ignore();
    | _ -> 
        begin
          if (v.vglob = true) then
            ignore()
          else
            numberVars := !numberVars + 1;
        end
    end;
    DoChildren;

  method vfunc(f: fundec) =
    Hashtbl.add tableFuncLocals f.svar.vname f.slocals; (*new*)
    Hashtbl.add tableFuncNumArgs f.svar.vname (List.length f.sformals);
    Hashtbl.add tableFuncFmls f.svar.vname f.sformals;

    List.iter self#doGlobal (List.map (fun x -> (f.svar.vname ^ "$" ^ (string_of_int x))) (self#createIntList 1 (List.length f.sformals)));
    self#doGlobal (f.svar.vname ^ "$return");
    DoChildren;

end
(* End of counter visitor *)



exception Non_struct of string;;
(*Some debugging methods*)
 let printCilObj d_obj = fun obj chan ->
   let d = dprintf "%a" d_obj obj in
   let s = sprint 80 d in
   ignore( fprintf chan "[%s]\n" s )

let printCilExp = printCilObj d_exp
and printCilStmt = printCilObj d_stmt
and printCilLval = printCilObj d_lval

class wpdsVisitor theFile ?outChan:(pdsout=stdout) ?errChan:(pdserr=stderr) () = object(self)
  inherit nopCilVisitor
      
      (* Name of current function being analyzed *)
  val mutable fname = "";
    (* Where to begin prestar from *)
  val mutable mainRetId = 0;
    (* The cilPrinter used by wpdsVisitor to output the WPDS *)
  val mutable wpdsPrinter = plainCilPrinter;

  method getPrestarId() : int = mainRetId

  method getCompInfoForType(t:typ) =
    match t with
    | TVoid(_)
    | TInt(_,_)
    | TFloat(_,_) ->
	begin
          raise (Non_struct "?.getCompInfoForType");
	end;
    | TPtr(t1,_)
    | TArray(t1,_,_) ->
	begin
          self#getCompInfoForType t1;
	end;
    | TComp(cinfo,attrs) ->
	begin
          cinfo;
	end;
    | TNamed(tinfo,attrs) ->
	begin
          self#getCompInfoForType tinfo.ttype;
	end;
    | _ ->
	begin
	  ignore (bug "[getCompInfoForType] not yet completed...\n");
	  failwith "internal error 1";
	end;

  method getCompInfoForLval(lv:lval) =
    match lv with
    | (Var(vinfo),_) ->
	begin
          self#getCompInfoForType vinfo.vtype;
	end;
    | (Mem(e),off) ->
	begin
          self#getCompInfoForExp e;
	end;

  method getCompInfoForExp(e:exp) =
    match e with
    | UnOp(_,_,t) | BinOp(_,_,_,t) | SizeOf(t) | AlignOf(t) | CastE(t,_) ->
	begin
          self#getCompInfoForType t;
	end;
    | Lval(lv) ->
	begin
          self#getCompInfoForLval lv;
	end;
    | _ ->
	begin
	  ignore (bug "[getCompInfoForExp] not yet completed...\n");
	  printCilExp e pdserr;
	  failwith "internal error 2";
	end;

  method getCompInfoForLhost(host:lhost) =
    match host with
    | Var(vinfo) ->
	begin
          self#getCompInfoForType vinfo.vtype;
	end;
    | Mem(e1) ->
	begin
          self#getCompInfoForExp e1;
	end;

  method findVarsInExp(e:exp)(b:bool) =
    match e with
    | Const(_) | SizeOf(_) | SizeOfStr(_) | AlignOf(_) -> 
	begin 
          ignore();	
	end;
    | SizeOfE(e1) | AlignOfE(e1) | UnOp(_,e1,_) | CastE(_,e1) ->
	begin
          ignore();
	end;
    | BinOp(op,e1,e2,t) ->
	begin
          match op with
          | Lt -> ignore();
          | Gt -> ignore();
          | Le -> ignore();
          | Ge -> ignore();
          | Eq -> ignore();
          | Ne -> ignore();
          | _ -> 
              self#findVarsInExp e1 true;
              self#findVarsInExp e2 true;
	end;
    | Lval(lv) ->
	begin
          match lv with
          | (host,Field(finfo,foff)) ->
              begin
		ignore(); (*for now*)
              end;
          | (Mem(e1),off) ->
              begin
		match e1 with
		| Lval(Var(vinfo),off) -> self#printSetArt vinfo.vname "OK";
		| _ -> ignore();
              end;
          | (Var(vinfo),off) ->
              begin
		self#printSetArt vinfo.vname "OK";
              end
	end;
    | AddrOf(lv) ->
	begin
          ignore(); (*for now*)
          (*self#doLval lv;*)
          (*match lv with
             | (Var(vinfo), off) ->
	     ignore( fprintf pdsout "\"%s\"" vinfo.vname );
             | _ ->  ignore( fprintf pdsout "[ERROR] New error cindy...\n");
           *)
	end;
    | StartOf(lv) ->
        ignore(); (*for now*)



	(* Check this*)
  method doLval( lv: lval ) =
    begin
      match lv with
      | (host,Field(finfo,foff)) ->
	  begin
            begin try
              let cinfo = self#getCompInfoForLhost host in
              begin
		(*ignore( fprintf pdsout "-,-,%s,%s:" cinfo.cname finfo.fname );*) (*Commented for now!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*)
		ignore( fprintf pdsout "\"%s\"" cinfo.cname );
              end;
            with Non_struct ns -> ignore( fprintf pdserr "[Non_struct] %s\n" ns )
            end;
	  end;
      | (Mem(e1),off) ->
	  begin
            self#doExp e1;
	  end;
      | (Var(vinfo),off) ->
	  begin
            ignore( fprintf pdsout "\"%s\"" vinfo.vname);
	  end
    end;


  method isTrusted(e:exp) =
    match e with
    | Const(CInt64(-67737868L,_,_)) -> true
    | Const(CInt64 _) -> false
    | Lval(_) -> false
    | _ -> 
	begin
          ignore( fprintf pdsout "\"unimplemented\"");
          false
	end;
	

(*5, 6 errors.txt*)
  method doExp(e:exp) =
    match e with
    | Const (CInt64 (value, _, _)) ->
	let loc =
	  match Error.byValue value with
	  | Some code -> Error.name code
	  | None -> "OK"
	in
	ignore (fprintf pdsout "\"%s\"" loc)
    | Const(_) | SizeOf(_) | SizeOfStr(_) | AlignOf(_) -> 
	begin 
          ignore( fprintf pdsout "\t\t\t<unimplemented reason=\"Expression: SizeOf, SizeOfStr and AlignOf\"/>\n");	
	end;
    | SizeOfE(e1) | AlignOfE(e1) | UnOp(_,e1,_) (*| CastE(_,e1)*) ->
	begin
          (*self#doExp e1;*)
          ignore( fprintf pdsout "\t\t\t<unimplemented reason=\"Expression: SizeOfE, AlignOfE and UnOp\"/>\n");		     
	end;
    | CastE(_,e1) ->
        self#doExp e1;
    | BinOp(_,e1,e2,t) ->
	begin
          (*self#doExp e1;
             self#doExp e2;*)
          ignore( fprintf pdsout "\t\t\t<unimplemented reason=\"Expression: BinOp\"/>\n");
	end;
    | Lval(lv) ->
	begin
          self#doLval lv;
	end

    | AddrOf(lv) ->
	begin
          self#doLval lv;
          (*match lv with
             | (Var(vinfo), off) ->
	     ignore( fprintf pdsout "\"%s\"" vinfo.vname );
             | _ ->  ignore( fprintf pdsout "[ERROR] New error cindy...\n");
           *)
	end
    | StartOf(lv) ->
        ignore( fprintf pdsout "\t\t\t<unimplemented reason=\"Expression: StartOf\"/>\n");
	

  method makeWeight (s:stmt) =
    let arithmetic = Arithmetic.assumeNonError pdsout s in

    match s.skind with
    | Instr [Set (receiverLval, senderExpr, location)] ->
	let receiver = Rules.getReceiver receiverLval in
	let sender = Rules.getSender senderExpr in
	begin
	  match receiver, sender with
	  | Some receiver, Some Rules.OK
	    when Hashtbl.mem arithmetic receiver
	    ->
	      ()

	  | Some receiver, Some sender ->
	      let trusted =
		if isInteger senderExpr = Some (-67737868L) then
		  Rules.Trusted
		else
		  Rules.Untrusted
	      in
	      Rules.set pdsout receiver sender trusted
	  | Some _, None ->
	      Rules.unimplemented pdsout "assignment with unusual sender"
          | None, Some _ ->
	      Rules.unimplemented pdsout "assignment with unusual receiver"
	  | None, None ->
	      Rules.unimplemented pdsout "assignment with unusual sender and receiver"
	end

    | Instr[Call _] ->
	Errormsg.s (Errormsg.bug "unexpected call instruction")

    | Instr[Asm(_,_,receivers,_,clobbers,_)] ->
	Rules.unimplemented pdsout "Instruction: Asm"

    | Instr [] | Block _ | Goto(_,_) | Return( None,_) | Loop _ ->
	begin
          ignore( fprintf pdsout"\t\t\t<identity/>\n" );
	end;
    | Instr _ ->
	begin
          ignore (bug "[makeWeight] Instr not atomized?\n");
          failwith "[makeWeight] internal error 3";
	end;
    | If( e,_,_,loc) ->
	begin
          (*self#findVarsInExp e;*)  (*Commented!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!****************************************************)
          ignore( fprintf pdsout"\t\t\t<identity/>\n" ); (*in case no artificial sets were created*)
	end;
    | Return( Some e, loc) ->
	Rules.unimplemented pdsout "Return-"
    | _ ->
	begin
          ignore (bug "[makeWeight] ???\n");
(*
   ignore( bug "[makeWeight] Matched '_'... Not yet implemented?" );
   ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);
   failwith "[makeWeight] internal error 4";
 *)
	end



  method writeBack(sender:string)(receiver:varinfo) =
    ignore( fprintf pdsout "\n\t\t\t<set loc=\"%s\"" sender);   
    ignore( fprintf pdsout " trusted=\"true\">\n");

    ignore( fprintf pdsout "\t\t\t\t<from loc=\"%s\"" receiver.vname);   
    ignore( fprintf pdsout "/>\n");
    ignore( fprintf pdsout "\t\t\t</set>\n"); 


  method writeBack99 sender receiver =
    Rules.set pdsout receiver sender Rules.Trusted


  method writeBack2(sender:string)(receiver:exp) =
    match receiver with
    | AddrOf(lv) ->
        begin
          ignore( fprintf pdsout "\n\t\t\t<set loc=");   
          self#doExp receiver;
          ignore( fprintf pdsout " trusted=\"true\">\n");

          ignore( fprintf pdsout "\t\t\t\t<from loc=\"%s\"/>" sender);   
          ignore( fprintf pdsout "\n\t\t\t</set>\n"); 
        end
    | _ ->  ignore( fprintf pdsout"\t\t\t<identity/>\n" ); (*ignore(); before*)

	

  method createIntList(initial:int)(length:int) =
    if (length <= 0) then
      []
    else
      initial::(self#createIntList (initial + 1) (length - 1));


  method getFirstN (args:exp list)(n:int) =
    if (n <= 0) then
      []
    else
      (hd args)::(self#getFirstN (tl args) (n - 1));


  method printSetArt(receiver:string)(sender:string) =
    ignore( fprintf pdsout "\t\t\t<set loc=\"%s\"" receiver);   
    ignore( fprintf pdsout " trusted=\"true\">\n");
    ignore( fprintf pdsout "\t\t\t\t<from loc=\"%s\"" sender);
    ignore( fprintf pdsout "/>\n");
    ignore( fprintf pdsout "\t\t\t</set>\n");     


  method printSet(receiver:string)(sender:exp) =
    begin match sender with
    | Const(CInt64(_,_,_)) | Lval(_) | CastE(TInt(_,_), Lval(_)) (*| AddrOf((Var(_),_))*) ->
        if isLvalOK sender then (*redundant if not Lval*)
          begin 
            ignore( fprintf pdsout "\n\t\t\t<set loc=\"%s\"" receiver);   
            ignore( fprintf pdsout " trusted=\"true\">\n");

            ignore( fprintf pdsout "\t\t\t\t<from loc=");
            self#doExp sender;   
            ignore( fprintf pdsout "/>\n");
            ignore( fprintf pdsout "\t\t\t</set>\n");    
          end
        else
          ignore( fprintf pdsout "\n\t\t\t<unimplemented reason=\"Lval at printSet\"/>\n");

    | AddrOf(lv) ->
        if checkLval lv then
          begin 
            ignore( fprintf pdsout "\n\t\t\t<set loc=\"%s\"" receiver);   
            ignore( fprintf pdsout " trusted=\"true\">\n");

            ignore( fprintf pdsout "\t\t\t\t<from loc=");
            self#doExp sender;   
            ignore( fprintf pdsout "/>\n");
            ignore( fprintf pdsout "\t\t\t</set>\n");    
          end
        else
          ignore( fprintf pdsout "\n\t\t\t<unimplemented reason=\"AddrOf lv at printSet\"/>\n"); 


    | _ ->  (**new: we know we are returning something, but we don't know what, result likely to overwrite something**)
	begin 
          ignore( fprintf pdsout "\n\t\t\t<set loc=\"%s\"" receiver);   
          ignore( fprintf pdsout " trusted=\"true\">\n");
	  
          ignore( fprintf pdsout "\t\t\t\t<from loc=\"OK\"");
          ignore( fprintf pdsout "/>\n");
          ignore( fprintf pdsout "\t\t\t</set>\n");    
        end

        (*ignore( fprintf pdsout"\t\t\t<identity/>\n" );*) (*ignore();*)
    end;


  method printUntrustedAssg(receiver:varinfo)(sender:string)(trusted:bool) =
    match receiver.vtype with
    | TInt(_,_) -> 
        begin
          ignore( fprintf pdsout "\t\t\t<set loc=\"%s\"" receiver.vname); 
          if trusted = false then  
            ignore( fprintf pdsout " trusted=\"false\">\n")
          else
            ignore( fprintf pdsout " trusted=\"true\">\n");
          ignore( fprintf pdsout "\t\t\t\t<from loc=\"%s\"" sender);
          ignore( fprintf pdsout "/>\n");
          ignore( fprintf pdsout "\t\t\t</set>\n");
	end;
    | _ -> ignore();



  method equal(expr:exp)(var:varinfo) =
    match expr with
    | Lval(Var(vinfo), _)
    | CastE(_, Lval(Var(vinfo),_)) -> 
        if vinfo.vname = var.vname then
          true
        else
          false; 
    | _ -> false;
  


  method makeArtificialInitialIntraEdge(fmls:varinfo list)(line:int)(file:string) =
    try
      ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.first' to='p'" fname );
      ignore( fprintf pdsout " toStack1='%s.%d'>\n" fname 0);
      ignore( fprintf pdsout"\t\t<Weight>\n" );

      if fname = "main" then
	ignore( fprintf pdsout"\t\t\t<uninitialized/>\n" )
      else
	ignore( fprintf pdsout"\t\t\t<identityGlobals/>\n" );


      (* Assigning exchange variables to formals *)
      let numArgs = Hashtbl.find tableFuncNumArgs fname in
      
      if numArgs > 0 then
        begin
          let exchangeVars = List.map (fun x -> fname ^ "$" ^ (string_of_int x)) (self#createIntList 1 numArgs) in
          let argVars = List.map (fun x -> x.vname) fmls in
          List.iter2 self#printSetArt argVars exchangeVars;

	  if fname = "printk" then
	    begin
	      ignore( fprintf pdserr "PRINTK2 !!! \n");
	      ignore (List.map (
		      fun f -> self#printUntrustedAssg f "OK" true;) fmls;)
	    end

        end
      else
        ignore( fprintf pdsout"\t\t\t<identity/>\n" );

      ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
      ignore( fprintf pdsout"\t\t</Weight>\n" );
      ignore( fprintf pdsout"\t</Rule>\n" );
    with Not_found -> ignore( fprintf pdsout "[ERROR in makeArtificialInitialIntraEdge] Function %s was not found in table\n" fname);
    | Invalid_argument ns -> ignore(fprintf pdsout "[ERROR in makeArtificialInitialIntraEdge] Invalid argument %s\n" ns);


  method makeArtificialFinalIntraEdge(args:exp list)(s:stmt)(s1:stmt)(funName:string)(receiver:string)(line:int)(file:string) =
    try
      (* Writing pointer variables back to corresponding exchange variables *)
      (* We assume that pointer variables are not changed inside the function and that they are not aliases *)

      if (Hashtbl.mem tableFuncNumArgs funName) then
	begin
          ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d.after' to='p'" fname s.sid );
          ignore( fprintf pdsout " toStack1='%s.%d'>\n" fname s1.sid);
          ignore( fprintf pdsout"\t\t<Weight>\n" );

          let numArgs = Hashtbl.find tableFuncNumArgs funName in
          begin 
            if numArgs > 0 then
              begin
                let exchangeVars = List.map (fun x -> funName ^ "$" ^ (string_of_int x)) (self#createIntList 1 numArgs) in                
                (*List.iter2 self#writeBack2 exchangeVars (Hashtbl.find tableFuncAcls funName);*) (* e.g. actual = foo$1 *) (*original, before considering variable parameter list functions*)
                List.iter2 self#writeBack2 exchangeVars (self#getFirstN (Hashtbl.find tableFuncAcls funName) numArgs);
              end
            else
              ignore( fprintf pdsout"\t\t\t<identity/>\n" );

            if receiver <> "" then
              self#printSetArt receiver (funName ^ "$return");
          end;
          ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
          ignore( fprintf pdsout"\t\t</Weight>\n" );
          ignore( fprintf pdsout"\t</Rule>\n" );
	end
      else
	ignore();
    with Not_found -> ignore( fprintf pdsout "[ERROR in makeArtificialFinalIntraEdge**] The function %s was not found in the table\n" funName);
    | Invalid_argument ns -> ignore(fprintf pdsout "[ERROR in makeArtificialFinalIntraEdge] Invalid argument %s\n" ns);


  method makeIntraEdge(s:stmt)(s1:stmt)(line:int)(file:string) =
    ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d' to='p'" fname s.sid );
    ignore( fprintf pdsout " toStack1='%s.%d'>\n" fname s1.sid);
    ignore( fprintf pdsout"\t\t<Weight>\n" );
    ignore( self#makeWeight s );
    ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
    ignore( fprintf pdsout"\t\t</Weight>\n" );
    ignore( fprintf pdsout"\t</Rule>\n" );



  method makePushRule(s:stmt)(s1:stmt)(funName:string)(args:exp list)(line:int)(file:string) =
    try
      if (Hashtbl.mem tableFuncNumArgs funName) then
	begin
          ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d' to='p'" fname s.sid );
          ignore( fprintf pdsout " toStack1='%s.first'" funName);
          ignore( fprintf pdsout " toStack2='%s.%d.after'>\n" fname s.sid );
          ignore( fprintf pdsout"\t\t<Weight>" );
          ignore( Arithmetic.assumeNonError pdsout s );

          let numArgs = Hashtbl.find tableFuncNumArgs funName in
	  begin 
	    if numArgs <> 0 then
	      let exchangeVars = List.map (fun x -> funName ^ "$" ^ (string_of_int x)) (self#createIntList 1 numArgs) in
              begin
                if (numArgs <> (List.length args)) then
                  ignore(fprintf pdserr "Function %s is a variable parameter list function\n" funName);

                List.iter2 self#printSet exchangeVars (self#getFirstN args numArgs);
              end;
            else
              ignore( fprintf pdsout"\n\t\t\t<identity/>\n" ); 
          end;

          Hashtbl.add tableFuncAcls funName args;  (*Adding args to table*)
	  
	end

      else 
        (* The function definition is not available...*)
        begin 
          ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d' to='p'" fname s.sid );
          ignore( fprintf pdsout " toStack1='%s.%d'>\n" fname s1.sid);
          ignore( fprintf pdsout"\t\t<Weight>\n" );

          ignore( fprintf pdserr "!!!! The definition for function %s is not available\n" funName);
	  ignore( Arithmetic.assumeNonError pdsout s );
          ignore( fprintf pdsout "\t\t\t<identity/>\n" ); 

	  if funName = "printk" then
	    begin
	      ignore( fprintf pdserr "PRINTK !!! \n");
	      ignore (List.map (
	      fun a -> match a with
                        | Lval(Var(varinfo),_) -> self#printUntrustedAssg varinfo "OK" true;
	               | _ -> ignore()) 

               args;)
	    end
          
          (*Printing corresponding Pop Rule, otherwise the rest of the program will be lost! *)
          (*ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.first' to='p'>\n\t\t<Weight>" funName );*)
          (*ignore( fprintf pdsout "\n\t\t\t<identity/>\n" ); *)
        end;

      ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
      ignore( fprintf pdsout "\t\t</Weight>\n\t</Rule>\n" );
      
    with Not_found -> ignore( fprintf pdsout "[ERROR in makePushRule] Function %s was not found in table\n" funName);
    | Invalid_argument ns -> ignore(fprintf pdsout "[ERROR in makePushRule] Invalid argument %s\n" ns);


	(* This function will obviously not be called when a function's definition is not available *)
  method makePopRuleRet(s:stmt)(retExp:exp)(line:int)(file:string) =
    try
      ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d' to='p'>\n\t\t<Weight>" fname s.sid );
      let arithmetic = Arithmetic.assumeNonError pdsout s in
      self#printSet (fname ^ "$return") retExp;
      
      (* Assigning pointer variables back to their corresponding exchange variables? Check this out *) (********************************)
      let numArgs = Hashtbl.find tableFuncNumArgs fname in
      let exchangeVars = List.map (fun x -> fname ^ "$" ^ (string_of_int x)) (self#createIntList 1 numArgs) in
      begin
        List.iter2 self#writeBack exchangeVars (Hashtbl.find tableFuncFmls fname);
      end;

      (*new*)
      let lslocals = Hashtbl.find tableFuncLocals fname in
      let lsfmls   = Hashtbl.find tableFuncFmls   fname in
      let erase var =
	if not (Hashtbl.mem arithmetic var) then
	  let trusted = self#equal retExp var in
	  self#printUntrustedAssg var "OK" trusted
      in
      List.iter erase lslocals;
      List.iter erase lsfmls;
      (*new until here*)

      ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
      ignore( fprintf pdsout "\t\t</Weight>\n\t</Rule>\n" );
      
    with Not_found -> ignore( fprintf pdsout "[ERROR in makePopRuleRet] Function %s was not found in table\n" fname);
    | Invalid_argument ns -> ignore(fprintf pdsout "[ERROR in makePopRuleRet] Invalid argument %s\n" ns);


  method makePopRuleNoRet(s:stmt)(line:int)(file:string) =
    try 
      ignore( fprintf pdsout "\t<Rule from='p' fromStack='%s.%d' to='p'>\n\t\t<Weight>" fname s.sid );

      let numArgs = Hashtbl.find tableFuncNumArgs fname in
      if numArgs <> 0 then
        let exchangeVars = List.map (fun x -> fname ^ "$" ^ (string_of_int x)) (self#createIntList 1 numArgs) in
        begin
          List.iter2 self#writeBack exchangeVars (Hashtbl.find tableFuncFmls fname);
        end;
      else
        ignore( fprintf pdsout"\n\t\t\t<identity/>\n" );

      let lslocals = Hashtbl.find tableFuncLocals fname in
      List.iter (fun x -> self#printUntrustedAssg x "OK" false) lslocals;	     
      let lsfmls = Hashtbl.find tableFuncFmls fname in
      List.iter (fun x -> self#printUntrustedAssg x "OK" false) lsfmls;	     

      ignore( fprintf pdsout "\t\t\t<source line=\"%d\" file=\"%s\"/>\n" line file );
      ignore( fprintf pdsout "\t\t</Weight>\n\t</Rule>\n" );


      
      (*Hashtbl.remove tableFuncAcls fname;*) (*Removing args from table*)
    with Not_found -> ignore(fprintf pdsout "[ERROR in makePopRuleRet] Function %s was not found in table\n" fname);
    | Invalid_argument ns -> ignore(fprintf pdsout "[ERROR in makePopRuleNORet] Invalid argument %s\n" ns);


	(*** Check this ***)
  method handleStmt (s:stmt) = fun (s1:stmt) -> 
    begin 
      match s.skind with
      | Instr[Set(_, _, loc)] 
      | Instr[Asm(_, _, _, _, _, loc)] 
	  (*| Instr [] *)
      | If(_, _, _, loc) | Loop(_, loc, _, _) | Goto (_, loc) ->
	  begin
            self#makeIntraEdge s s1 loc.line loc.file;
	  end
      | Instr[] -> 
	  begin
            self#makeIntraEdge s s1 0 "";  (*new*)
	  end;

      | Block _ ->
	  begin
            (* I put block into a separate case ... *)
            self#makeIntraEdge s s1 0 "";  (*I don't know why this gets a weight and I don't believe it will ever be different to identity*)
	  end
      | Instr[Call(Some receiver, f, args, loc)] ->
	  begin 
            match f with
            | Lval(Var(info),oset) ->
		begin
                  self#makePushRule s s1 info.vname args loc.line loc.file;
                  (* make artifical final intra edge here, otherwise I won't be able to have the receiver's info*)
                  (match receiver with
                  | (Var(cvi),NoOffset) -> 
                      self#makeArtificialFinalIntraEdge args s s1 info.vname cvi.vname 0 "";
                  | _ -> (*failwith "internal error 5";*)  ignore();   (*For now!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*)
                  );
		end;
            | Lval(Mem(memExp),oset) ->

		(*I modified the code from here...*)
		begin try
                  let ls = Ptranal.resolve_funptr memExp in
                  List.iter (fun f -> self#makePushRule s s1 f.svar.vname args loc.line loc.file;
                    (match receiver with
                    | (Var(cvi),NoOffset) -> 
                        self#makeArtificialFinalIntraEdge args s s1 f.svar.vname cvi.vname 0 "";
                    | _ -> failwith "internal error 5a";
                    );
                            ) ls;

		with Not_found ->
                  match memExp with
                  | Lval(Var(vinfo), ooset) -> ignore (fprintf pdserr "****** Pointer to a function. Name: %s File: %s Line: %d\n" vinfo.vname loc.file loc.line);
                  | _ ->  ignore (fprintf pdserr "**--**--** Pointer to a function. I don't know wich function...\n");
		end;
		(* till here*)


            | _ ->
		begin
                  ignore (bug "Call is not an Lval?\n");
                  ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);
                  failwith "internal error 6";
		end;
	  end
      | Instr[Call(None, f, args, loc)] ->
	  begin 
            match f with
            | Lval(Var(info),oset) ->
		begin
                  self#makePushRule s s1 info.vname args loc.line loc.file;
                  (* make artifical final intra edge here, otherwise I won't be able to have the receiver's info*)
                  self#makeArtificialFinalIntraEdge args s s1 info.vname "" 0 "";
		end;
            | Lval(Mem(memExp),oset) ->

		(*I modified the code from here...*)
		begin try
                  let ls = Ptranal.resolve_funptr memExp in
                  List.iter (fun f -> self#makePushRule s s1 f.svar.vname args loc.line loc.file;
                    self#makeArtificialFinalIntraEdge args s s1 f.svar.vname "" 0 "";
                            ) ls;

		with Not_found ->
                  match memExp with
                  | Lval(Var(vinfo), ooset) -> ignore (fprintf pdserr "****** Pointer to a void function. Name: %s File: %s Line: %d\n" vinfo.vname loc.file loc.line);
                  | _ ->  ignore (fprintf pdserr "**--**--** Pointer to a void function. I don't know wich function...\n");
		end;
		(* till here*)
            | _ ->
		begin
                  ignore (bug "Call is not an Lval?\n");
                  ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);
                  failwith "internal error 7";
		end;
	  end

      | Instr _ ->
	  begin
            ignore (bug "[handleStmt] Instr should have been atomized");
            ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);
            failwith "internal error 8";
	  end
      | _ ->
	  begin
            ignore( fprintf pdserr "\n\tOther stmt\n");
            self#makeIntraEdge s s1 0 "";  
            ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);
            failwith "internal error 9";
	  end
    end;



  method vstmt(s: stmt) = begin
    match s.skind with
      (* Return handled separately b/c it has no successors *)
    | Return( Some(retExp), loc ) ->
	if fname == "main" then
          ( mainRetId <- s.sid );
	self#makePopRuleRet s retExp loc.line loc.file
    | Return( None, loc ) ->
	if fname == "main" then
          ( mainRetId <- s.sid );
	self#makePopRuleNoRet s loc.line loc.file
    | _ ->
	if (List.length s.succs) > 0 then
          List.iter (self#handleStmt s) s.succs
	else
          begin
            ignore(fprintf pdserr "++ Created a pop rule for a statement other than a 'Return' statement which does not have any successors");
	    self#makePopRuleNoRet s 0 "";
          end
	    
  end;
    DoChildren;


  method vfunc(f: fundec) =
    (*
       This atomizes the statements so that we only ever look
       at them 1 at a time. This makes creating the WPDS easier as
       Call instructions have been pulled out of blocks. However, it
       might make more sense at some time to simply split Blocks at 
       Call Instr...
     *)

    IsolateInstructions.visit f; (*Cindy*)

    (*
       Create the CFG.
       - Cfg.computeFileCFG is not used b/c it does not do the simplification
       prepareCfg does.
     *)

    prepareCFG f;
    computeCFGInfo f false;
    
    (*For debugging purposes*)
    (*    
       ignore(fprintf pdserr "FOR DEBUGGING\n";);
       ignore(fprintf pdserr "--Name of function: %s\n" f.svar.vname);
       List.iter (fun s -> ignore( fprintf stderr "%d -- " s.sid); printCilStmt s stderr) f.sallstmts;
     *)

    (*new*)
    (*List.iter (fun v -> ignore(fprintf pdserr "%s #### %s: " f.svar.vname v.vname)) f.slocals;*)
    (*new until here*)

    fname <- f.svar.vname;
    self#makeArtificialInitialIntraEdge f.sformals 0 "";
    DoChildren

end


(* This flag determines whether the generated WPDS Query will do prestar.
   It controls 2 output strings in feature.fd_doit *)
let doPrestar = ref false
let pdsName = ref ""
let dtd = ref None
let functionfile = ref None

(* So output channels can be specified on the cmd line.
   --Note: only pdsout can be changed currently. *)
let pdsout = ref stdout
let pdserr = ref stderr


let rec insertBreaks calls =
  if List.length calls <> 0 then
    (hd calls)::(mkStmt (Break(locUnknown)))::(insertBreaks (tl calls))
  else
    []

(* Returns true is a main function is not found *)
let isMainNeeded file =
  let funcNames = List.map (fun g -> match g with 
  | GFun(fdec, loc) ->  fdec.svar.vname;
  | _ -> ""; ) file.globals in
  try 
    begin 
      ignore(List.find (fun name -> name = "main") funcNames);
      false;
    end;
  with
    Not_found -> true

	
(* Returns a statement consisting of a call to the given callee function *)
let createCallStmt file caller counter callee =
  try
    let global = List.find (( fun n g -> match g with 
    | GFun(fdec, loc) -> 
        if fdec.svar.vname = n then true else false;
    | _ -> false;
                             ) callee ) file.globals in
    match global with
    | GFun(fundec, loc) ->
        begin
          let ls_types = List.map (fun vinfo -> vinfo.vtype) fundec.sformals in
          let ls_actuals = List.map (fun t -> Lval(var (makeTempVar caller t))) ls_types in
          
          let call_instr = Call(None, Lval(var fundec.svar), ls_actuals, locUnknown) in
          let call_stmt = mkStmtOneInstr call_instr in
          begin
            call_stmt.labels <- [Case(integer !counter, locUnknown)];
            counter := !counter + 1;
            call_stmt;
          end;
        end;
    | _ ->  failwith ("[ERROR] " ^ callee ^ " is not a function")
	  
  with
  | Not_found -> failwith ("[ERROR] Cannot analyze " ^ callee ^ ": no implementation found")


(* Creates an artificial function funcName whose body consists of a list of calls to each 
   of the functions listed in functionsToCall *)
let createArtificalMain file funcName functionsToCall =
  let newFundec = Cil.emptyFunction funcName in

  (* Creating statements *)
  let counter = ref 0 in

  let default_case = mkEmptyStmt() in
  default_case.labels <- [ Default locUnknown ];
  let switch_stmt_list = ( List.map (createCallStmt file newFundec counter) functionsToCall ) @ [ default_case ] in
  let switch_block = mkBlock (insertBreaks switch_stmt_list) in
  let switch_exp = Lval(var(makeTempVar newFundec intType)) in
  let switch_stmt = mkStmt (Switch(switch_exp, switch_block, switch_stmt_list, locUnknown)) in

  let while_stmt = mkStmt (Loop(mkBlock [switch_stmt], locUnknown, None, None)) in
  let return_stmt = mkStmt (Return(None, locUnknown)) in

  (* Creating a block containing the statements *)
  let body = mkBlock [while_stmt; return_stmt] in             
  begin
    newFundec.sbody <- body;
    GFun(newFundec, locUnknown);
  end


let rec read_lines channel =
  try
    let line = input_line channel in
    line::(read_lines channel)
  with End_of_file -> close_in channel; [] 


let getFunctions filename =
  match !filename with
  | Some path ->  read_lines (open_in path);
  | None -> []




(* TODO: get rid of this xml somehow. this is no longer generic code *)
let feature : featureDescr = 
  { fd_name = "wpds";
    fd_enabled = ref false;
    fd_description = "save WPDS description and query (see --queryfile)";
    fd_extraopt = [
    ("--queryfile",
     Arg.String
       (fun s ->
         try let c = open_out s in pdsName :=s ; pdsout := c
         with Sys_error sc -> ignore( "Cannot open '%s'\n", s ) ),
     "<fname>: Name of file to write WPDS Query XML file");
    ("--poststar",Arg.Unit (fun _ -> doPrestar := false),
     "Generate a poststar query (prestar is default)");
    ("--dtd", Arg.String (fun path -> dtd := Some path),
     "Name the given DTD in a DOCTYPE SYSTEM declaration in the WPDS XML file");
    ("--functionfile", Arg.String (fun path -> functionfile := Some path),
     "File that contains the names of the functions to be called from the automatically generated main function");
    ("--error-codes", Arg.String (fun path -> Error.initialize path),
     "File containing error code names and values");
  ];
    fd_doit = 
    (function (f: file) ->
      begin
        begin
	  match !dtd with
	  | None -> ()
	  | Some path ->
	      let escaped =
		let buffer = Buffer.create (String.length path) in
		let escape = function
		  | '<' | '>' | '&' | '"' as dangerous ->
		      Printf.bprintf buffer "&#%d;" (int_of_char dangerous)
		  | safe ->
		      Buffer.add_char buffer safe
		in
		String.iter escape path;
		Buffer.contents buffer
	      in
	      ignore( fprintf !pdsout "<!DOCTYPE Query SYSTEM \"%s\">\n" escaped )
	end;


        (*Pointer analysis*)	    
        (*Ptranal.analyze_mono := false;*) (*Enables context sensitivity in Golf analysis*)
        (*Ptranal.analyze_file f;*)
        (*Ptranal.compute_results true;*)

	(********************Creating articial function main************************************)

        if isMainNeeded f then
          begin
            let functionsToCall = getFunctions functionfile in
            let newGlobal = createArtificalMain f "main" functionsToCall in
            f.globals <- f.globals@[newGlobal];
            ignore(fprintf !pdserr "+ Added an artificial main function\n");
          end;


	(********************Done creating articial function main************************************)

        (************************)
        connectFunctionPointers f;
        (************************)

        let vis0 = new sourceVisitor f ~outChan:!pdsout ~errChan:!pdserr ()
        and vis1 = new counterVisitor f ~outChan:!pdsout ~errChan:!pdserr ()
        and vis2 = new wpdsVisitor f ~outChan:!pdsout ~errChan:!pdserr ()
        and query = if !doPrestar then "prestar" else "poststar" in
        begin
          visitCilFileSameGlobals (vis0 :> cilVisitor) f;
          Conditionals.condVisitor !pdsout f;

	  ignore (fprintf !pdsout "<Query type='%s'>\n" query);
	  ignore (fprintf !pdsout "\t<WPDS>\n");
	  ignore (fprintf !pdsout "\t<Globals>\n");
          visitCilFileSameGlobals (vis1 :> cilVisitor) f;(******************************************************)
            iterGlobals f (fun g ->  match g with
            | GVar(vinfo,_,_) -> 
                begin
                  ignore (fprintf !pdsout "\t\t<var name=\"%s\"/>\n" vinfo.vname);
                  numberVars := !numberVars + 1;
                  numberGlobalVars := !numberGlobalVars + 1;
                end;
            | _ -> ignore()
                          );


	  ignore (fprintf !pdsout "\t</Globals>\n");
          ignore( fprintf !pdsout "\t<Variables globals=\"%d\" locals=\"%d\"/>\n" !numberGlobalVars (!numberVars - !numberGlobalVars));

          visitCilFileSameGlobals (vis2 :> cilVisitor) f;
          ignore( fprintf !pdsout "\t</WPDS>\n");
	  
          ignore( fprintf !pdsout "\t<WFA type='INORDER'>\n" );
          ignore( fprintf !pdsout "\t\t<State Name='p' initial='TRUE'><Weight><zero/></Weight></State>\n" );
          ignore( fprintf !pdsout "\t\t<State Name='accept' final='TRUE'><Weight><zero/></Weight></State>\n" );


          let mainId = if !doPrestar then vis2#getPrestarId() else 0 in
          ignore( fprintf !pdsout "\t\t<Trans from='p' stack='main.%d' to='accept'><Weight><one/></Weight></Trans>\n" mainId ); (*Uncomment this later!!!!!!!!!!!!!!!!!1*)
          (*ignore( fprintf !pdsout "\t\t<Trans from='p' stack='ext2_delete_inode.first' to='accept'><Weight><one/></Weight></Trans>\n" );*)

          ignore( fprintf !pdsout "\t</WFA>\n" );
          ignore( fprintf !pdsout "</Query>\n" );
	  
        end;
        (if !pdsName <> "" then
          try close_out !pdsout
          with Sys_error sc -> ignore( fprintf !pdserr "Cannot close '%s'\n" !pdsName));
        
      end;
    );
    fd_post_check = true;
  } 


(*****Checking whether instructions are isolated*****)
(*print_string "before prepareCFG.. \n";
   CheckIsolated.checkIsolated theFile; 
   print_string "after prepareCFG...\n";*)
(*delete*) (*Rmtmps.removeUnusedTemps theFile;*)
(*delete*) (*Deadcodeelim.elim_dead_code f;*)
(*delete*) (*Cfg.cfgFun f;*)
(*delete*) (*ignore(Cil.dumpStmt wpdsPrinter pdserr 0 s);*)
